# cython: profile=False
cimport numpy as np

import numpy as np


cdef HEAP_DATA MASK = 0xffffffff00000000
cdef int REMOVED = -1


cdef HEAP_DATA __heap_value(double dist, int i):
    """Embed a distance and an integer as a single int64 value."""
    # distances are positive float values
    cdef HEAP_DATA value = (<HEAP_DATA *> &dist)[0]
    # zero out lower 32 bits, embed integer id in mantissa
    value &= MASK
    value |= i
    return value


cdef double __get_distance(HEAP_DATA value):
    """Get the distance part of a value."""
    cdef HEAP_DATA data = value & MASK
    return (<double *> &data)[0]


cdef int __get_id(HEAP_DATA value):
    """Get the id part of a value."""
    return value & ~MASK


cdef void __swap(HEAP_DATA[::1] l, int[::1] ids, int i, int j):
    """Swap two values in the heap."""
    l[i], l[j] = l[j], l[i]
    # update id : index in array mapping
    ids[__get_id(l[i])], ids[__get_id(l[j])] = i, j


cdef void __siftup(HEAP_DATA[::1] l, int[::1] ids, int i):
    """Bubble a value upwards."""
    # O(log n)
    while (i >> 1) > 0:
        parent = i >> 1
        # heap invariant broken, fix
        if l[i] > l[parent]:
            __swap(l, ids, i, parent)
            i = parent
        # heap invariant mantained
        else:
            break


cdef void __siftdown(HEAP_DATA[::1] l, int[::1] ids, int i):
    """Bubble a value downwards."""
    # O(log n)
    size = l.shape[0] - 1
    while (i << 1) <= size:
        left, right = i << 1, (i << 1) | 1
        # pick larger child
        child = left if right > size or l[left] > l[right] else right
        # heap invariant broken, fix
        if l[child] > l[i]:
            __swap(l, ids, i, child)
            i = child
        # heap invariant mantained
        else:
            break


cdef void __heapify(HEAP_DATA[::1] l, int[::1] ids):
    """Ensure heap property of data."""
    # O(n)
    for i in range((l.shape[0] - 1) >> 1, 0, -1):
        __siftdown(l, ids, i)


def resize(array):
    """Double the size of the array."""
    return np.append(array, np.zeros(len(array), np.asarray(array).dtype))


cdef class Heap:

    """Maximum heap which stores ids to support update keys."""

    def __init__(self, double[::1] dists, long[::1] ids) -> None:
        """Construct a new heap from the given data."""
        cdef int i
        self.size = dists.shape[0]
        # dummy value at the front
        self.l = np.zeros(self.size + 1, dtype=np.ulonglong)
        for i in range(self.size):
            self.l[i + 1] = __heap_value(dists[i], ids[i])
        length = max(np.max(ids) + 1 if ids.shape[0] > 0 else 1, 1)
        self.ids = np.zeros(length, dtype=np.int32)
        for i in range(self.size):
            self.ids[ids[i]] = i + 1
        # enforce heap ordering
        __heapify(self.l, self.ids)

    def __len__(self) -> int:
        """Length of the heap."""
        return self.size

    def __str__(self) -> str:
        """Human-readable representation of the heap."""
        return str(
            [
                (__get_distance(x), __get_id(x))
                for x in np.asarray(self.l[1: 1 + self.size])
            ]
        )

    def __repr__(self) -> str:
        """String representation of the heap."""
        return f"Heap({str(self)})"

    def __getitem__(self, int i) -> HEAP_DATA:
        """Get the i-th item of the heap."""
        return self.l[i + 1]

    def push(self, double dist, int i) -> None:
        """Add a value to the heap."""
        self.size += 1
        # hit capacity, resize
        if self.size == self.l.shape[0]:
            self.l, self.ids = resize(self.l), resize(self.ids)
        # add data at the end
        self.l[self.size] = __heap_value(dist, i)
        self.ids[i] = self.size
        # restore heap property
        __siftup(self.l, self.ids, self.size)

    cdef HEAP_DATA __pop(self):
        """Remove a value from the heap."""
        cdef HEAP_DATA value = self.l[1]
        # replace with last leaf
        __swap(self.l, self.ids, 1, self.size)
        self.l[self.size] = 0
        self.ids[__get_id(value)] = REMOVED
        self.size -= 1
        # restore heap property
        __siftdown(self.l, self.ids, 1)
        return value

    def pop(self) -> tuple:
        """Remove a value from the heap."""
        cdef HEAP_DATA value = self.__pop()
        return __get_distance(value), __get_id(value)

    cdef int __update_key(self, int i, double dist):
        """Update the value of key."""
        cdef:
            int k
            float value

        k = self.ids[i]
        if k == REMOVED:
            return 0
        value = __get_distance(self.l[k])
        # overwrite value
        self.l[k] = __heap_value(dist, i)
        # restore heap property
        if dist < value:
            __siftdown(self.l, self.ids, k)
        if dist > value:
            __siftup(self.l, self.ids, k)
        return 1

    def update_key(self, int i, double dist) -> bool:
        """Update the value of key."""
        return bool(self.__update_key(i, dist))

    cdef int __decrease_key(self, int i, double dist):
        """Decrease the value of key."""
        cdef:
            int k
            float value

        k = self.ids[i]
        if k == REMOVED:
            return 0
        value = __get_distance(self.l[k])
        # only overwrite value if less
        if dist < value:
            self.l[k] = __heap_value(dist, i)
            # restore heap property
            __siftdown(self.l, self.ids, k)
        return 1

    def decrease_key(self, int i, double dist) -> bool:
        """Decrease the value of key."""
        return bool(self.__decrease_key(i, dist))



ctypedef int MKL_INT

cdef extern from "mkl.h":
    # CBLAS routines
    void cblas_daxpby(
        const MKL_INT N,
        const double alpha,
        const double *X,
        const MKL_INT incX,
        const double beta,
        double *Y,
        const MKL_INT incY
    )

    # vector math library (VML) routines
    void vdSqrt(const MKL_INT n, const double a[], double r[])
    void vdExp(const MKL_INT n, const double a[], double r[])


from kolesky.ordering import p_reverse_maximin
from kolesky.ordering import sparsity_pattern

import numpy as np
from scipy.spatial import KDTree
import scipy.linalg
import scipy.sparse as sparse
import sklearn.gaussian_process.kernels as kernels


def __supernodes(sparsity, lengths, lamb):
    groups = []
    candidates = set(range(len(lengths)))
    agg_sparsity = {}
    i = 0
    while len(candidates) > 0:
        while i not in candidates:
            i += 1
        group = sorted(j for j in sparsity[i] if lengths[j] <= lamb * lengths[i] and j in candidates)
        groups.append(group)
        candidates -= set(group)
        s = sorted({k for j in group for k in sparsity[j]})
        agg_sparsity[group[0]] = s
        positions = {k: j for j, k in enumerate(s)}
        for j in group[1:]:
            agg_sparsity[j] = np.empty(len(s) - positions[j], dtype=int)
    return groups, agg_sparsity


def __cols(theta):
    return np.flip(np.linalg.cholesky(np.flip(theta))).T

def __aggregate_chol(points, kernel, sparsity, groups):
    n = len(points)
    ptr = np.cumsum([0] + [len(sparsity[i]) for i in range(n)])
    data, indices = np.zeros(ptr[-1]), np.zeros(ptr[-1])
    for group in groups:
        s = sorted(sparsity[group[0]])
        positions = {i: k for k, i in enumerate(s)}
        L_group = __cols(kernel(points[s]))
        for i in group:
            k = positions[i]
            e_k = np.zeros(len(s))
            e_k[k] = 1
            col = scipy.linalg.solve_triangular(L_group, e_k, lower=True, check_finite=False)
            data[ptr[i] : ptr[i + 1]] = col[k:]
            indices[ptr[i] : ptr[i + 1]] = s[k:]
    return sparse.csc_matrix((data, indices, ptr), shape=(n, n))

def kl_cholesky(points, kernel, rho, lamb, initial = None, p = 1):
    indices, lengths = p_reverse_maximin(points, initial, p)
    ordered_points = points[indices]
    sparsity = sparsity_pattern(ordered_points, lengths, rho)
    groups, agg_sparsity = __supernodes(sparsity, lengths, lamb)
    return __aggregate_chol(ordered_points, kernel, agg_sparsity, groups), indices



ctypedef unsigned long long HEAP_DATA

cdef double __get_distance(HEAP_DATA value)
cdef int __get_id(HEAP_DATA value)

cdef class Heap:

    cdef HEAP_DATA[::1] l
    cdef int[::1] ids
    cdef int size

    cdef HEAP_DATA __pop(self)
    cdef int __update_key(self, int i, double dist)
    cdef int __decrease_key(self, int i, double dist)



from .maxheap cimport Heap
from scipy.spatial import KDTree
import numpy as np
cimport numpy as np
from . cimport mkl

cpdef void update_dists(
    Heap heap,
    double[:, ::1] dists,
    double[::1] dists_k,
    long[::1] js,
):
    """Update the distance table and heap."""
    cdef:
        int p, index, i, insert
        long j
        float d

    p = dists.shape[1]
    for index in range(js.shape[0]):
        j = js[index]
        d = dists_k[index]
        # insert d into dists[j], pushing out the largest value
        i = 0
        for i in range(p):
            if d <= dists[j, i]:
                break
        insert = i
        for i in range(p - 1, insert, -1):
            dists[j, i] = dists[j, i - 1]
        if insert < p:
            dists[j, insert] = d
        heap.__decrease_key(j, dists[j, p - 1])

np.import_array()
cdef double[::1] _distance_vector(double[:, ::1] points, double[::1] point):
   cdef:
       int n, i, j
       double dist, d
       double *start
       double *p
       double[::1] dists
   n = points.shape[1]
   start = &points[0, 0]
   p = &point[0]
   dists = np.empty(points.shape[0], np.float64)
   for i in range(points.shape[0]):
       dist = 0
       for j in range(n):
           d = (start + i*n)[j] - p[j]
           dist += d*d
       dists[i] = dist
   mkl.vdSqrt(points.shape[0], &dists[0], &dists[0])
   return dists

cpdef tuple reverse_maximin(np.ndarray[np.float64_t, ndim=2] points, double[:, ::1] initial = None):
    cdef:
        int n, i, k, index, j, start
        double lk
        long[::1] indices
        double[::1] lengths
        Heap heap
        double[::1] dists
        list js
    n = points.shape[0]
    indices = np.empty(n, dtype = np.long)
    lengths = np.empty(n, dtype = np.float64)
    if initial is None or initial.shape[0] == 0:
        k = 0
        dists = _distance_vector(points, points[0])
        indices[n - 1] = k
        lengths[n - 1] = np.inf
        start = n - 2
    else:
        initial_tree = KDTree(initial)
        dists = initial_tree.query(points)[0]
        start = n - 1
    tree = KDTree(points)
    heap = Heap(dists, np.arange(n))
    for i in range(start, -1, -1):
        lk, k = heap.pop()
        indices[i] = k
        lengths[i] = lk
        js = tree.query_ball_point(points[k], lk)
        dists = _distance_vector(points[js], points[k])
        for index in range(len(js)):
            j = js[index]
            heap.decrease_key(j, dists[index])
    return indices, lengths

cpdef tuple p_reverse_maximin(np.ndarray[np.float64_t, ndim=2] points, double[:, ::1] initial = None, int p = 1):
    cdef:
        int n, i, k
        double inf, lk
        long[::1] indices
        double[::1] lengths
        double[:, ::1] dists
        double[::1] dists_k
        Heap heap
        list js
    inf = 1e6
    n = points.shape[0]
    indices = np.empty(n, dtype = np.long)
    lengths = np.empty(n, dtype = np.float64)
    if initial is None or initial.shape[0] == 0:
        dists = np.array([[-i + inf] * p for i in range(n)])
    else:
        initial_tree = KDTree(initial)
        dists = initial_tree.query(points, p)[0]
        dists = dists.reshape(n, p)
    tree = KDTree(points)
    heap = Heap(np.max(dists, axis=1), np.arange(n))
    for i in range(n - 1, -1, -1):
        lk, k = heap.pop()
        indices[i] = k
        if lk < inf - n:
            lengths[i] = lk
        else:
            lengths[i] = np.inf
        js = tree.query_ball_point(points[k], lk)
        dists_k = _distance_vector(points[js], points[k])
        update_dists(heap, dists, dists_k, np.array(js, dtype=np.int64))
    return indices, lengths

cpdef object[::1] sparsity_pattern(double[:, ::1] points, double[::1] lengths, double rho):
    cdef:
        int n, i, offset
        long j
        double length_scale
        object[::1] sparsity
    n = points.shape[0]
    tree = KDTree(points)
    offset = 0
    length_scale = lengths[0]
    sparsity = np.empty(n, dtype = object)
    for i in range(n):
        if lengths[i] > 2 * length_scale:
            tree = KDTree(points[i:])
            offset = i
            length_scale = lengths[i]
        sparsity[i] = [
            offset + j
            for j in tree.query_ball_point(points[i], rho * lengths[i])
            if offset + j >= i
        ]
    return sparsity


from pathlib import Path

import numpy as np
from Cython.Build import cythonize
from setuptools import Extension, setup, find_packages

path = Path(__file__).parent


include_dirs = [
    str(path / "include"),
    str(path / ".venv/include"),
    np.get_include(),
]
library_dirs = [
    str(path / "lib"),
    str(path / ".venv/lib"),
]
# https://www.hlibpro.com/doc/3.1/install.html
# https://www.intel.com/content/www/us/en/develop/documentation/get-started-with-mkl-for-dpcpp/top.html
libraries = [
    "lapack",
    "blas",
    "boost_filesystem",
    "boost_system",
    "boost_program_options",
    "boost_iostreams",
    "tbb",
    "z",
    "metis",
    "fftw3",
    "gsl",
    "gslcblas",
    "m",
    "hdf5",
    "hdf5_cpp",
    "m",
    "stdc++",
] + [
    "mkl_intel_lp64",
    "mkl_intel_thread",
    "mkl_core",
    "iomp5",
    "pthread",
    "m",
    "dl",
]

extra_compile_args = ["-Wall", "-O3", "-fopenmp", "-fPIC"]

extensions = [
    Extension(
        "*",
        ["kolesky/*.pyx"],
        define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
        include_dirs=include_dirs,
        library_dirs=library_dirs,
        libraries=libraries,
        extra_compile_args=extra_compile_args,
    ),
]

setup(
    name= 'kolesky',
    author = 'Raymond Zeng',
    version = '0.1.3',
    packages = find_packages(),
    long_description=open('README.md').read(),
    install_requires = ["numpy", "scipy", "scikit-learn", "mkl>=2021.4.0"],
    ext_modules=cythonize(
        extensions,
        annotate=True,
        compiler_directives={
            "language_level": 3,
            "boundscheck": False,
            "wraparound": False,
            "initializedcheck": False,
            "cdivision": True,
        },
    ),
)